import itertools
from copy import deepcopy
from functools import partial
from typing import Literal

import dash_mantine_components as dmc
from dash.development.base_component import Component
from dash_iconify import DashIconify

from dash_pydantic_form.form_section import FormSection
from dash_pydantic_form.ids import form_base_id
from dash_pydantic_utils import deep_merge

from .form_layout import FormLayout, Position


class AccordionFormLayout(FormLayout):
    """Accordion form layout.

    Parameters
    ----------
    render_kwargs: dict
        Kwargs to pass to the render method.
    sections: list[FormSection]
        List of form sections.
    remaining_fields_position: Position
        Position of the remaining fields, one of 'top', 'bottom' or 'none'.
    """

    class ids:
        """Accordion form ids."""

        accordion = partial(form_base_id, "_pydf-accordion")

    sections: list[FormSection]
    remaining_fields_position: Position = "top"
    layout: Literal["accordion"] = "accordion"

    def render(  # noqa: PLR0913
        self,
        *,
        field_inputs: dict[str, Component],
        aio_id: str,
        form_id: str,
        path: str,
        **_kwargs,
    ):
        """Render the sections in an accordion."""
        kwargs = deepcopy(self.render_kwargs)
        accordion_styles = deep_merge(
            {
                "control": {"padding": "0.5rem"},
                "label": {"padding": 0},
                "item": {
                    "border": "1px solid color-mix(in srgb, var(--mantine-color-gray-light), transparent 40%)",
                    "background": "color-mix(in srgb, var(--mantine-color-gray-light), transparent 80%)",
                    "marginBottom": "0.5rem",
                    "borderRadius": "0.25rem",
                },
                "content": {
                    "display": "flex",
                    "flexDirection": "column",
                    "gap": "0.375rem",
                    "padding": "0.125rem 0.5rem 0.5rem",
                },
            },
            kwargs.pop("styles", {}),
        )
        multiple = kwargs.pop("multiple", True)
        accordion = dmc.Accordion(
            [
                dmc.AccordionItem(
                    [
                        dmc.AccordionControl(
                            dmc.Text(
                                ([DashIconify(icon=section.icon)] if section.icon else []) + [section.name],
                                style={"display": "flex", "alignItems": "center", "gap": "0.5rem"},
                                fw=600,
                            ),
                        ),
                        dmc.AccordionPanel(
                            self.grid([field_inputs[field] for field in section.fields if field in field_inputs]),
                        ),
                    ],
                    value=section.name,
                )
                for section in self.sections
            ],
            value=[section.name for section in self.sections if section.default_open]
            if multiple
            else next((section.name for section in self.sections if section.default_open), None),
            styles=accordion_styles,
            id=self.ids.accordion(aio_id, form_id, path),
            multiple=multiple,
            **kwargs,
        )

        fields_not_in_sections = set(field_inputs) - set(itertools.chain.from_iterable(s.fields for s in self.sections))

        if self.remaining_fields_position == "none" or not fields_not_in_sections:
            return [accordion]
        if self.remaining_fields_position == "top":
            return [self.grid([v for k, v in field_inputs.items() if k in fields_not_in_sections], mb="sm"), accordion]
        return [accordion, self.grid([v for k, v in field_inputs.items() if k in fields_not_in_sections], mb="sm")]



import json
from abc import ABC, abstractmethod
from typing import Annotated, Literal, Union

import dash_mantine_components as dmc
from dash.development.base_component import Component
from plotly.io.json import to_json_plotly
from pydantic import BaseModel, Discriminator, Field, TypeAdapter, field_serializer

from dash_pydantic_utils import get_all_subclasses

Children_ = Component | str | int | float
Children = Children_ | list[Children_]
Position = Literal["top", "bottom", "none"]


class FormLayout(BaseModel, ABC):
    """Abstract form layout class, base for all form layouts.

    Parameters
    ----------
    render_kwargs: dict | None
        Kwargs to pass to the render method.
    """

    render_kwargs: dict = Field(default_factory=dict)
    layout: str

    @abstractmethod
    def render(  # noqa: PLR0913
        self,
        *,
        field_inputs: dict[str, Component],
        aio_id: str,
        form_id: str,
        path: str,
        read_only: bool | None,
        form_cols: int,
    ):
        """Render the form layout."""
        raise NotImplementedError

    def model_post_init(self, _context):
        """Model post init."""
        if self.render_kwargs is None:
            self.render_kwargs = {}

    @field_serializer("render_kwargs")
    def serialize_render_kwargs(self, value):
        """Serialize render kwargs, allowing Dash object values."""
        jsonified = to_json_plotly(value)
        if not isinstance(jsonified, str):
            raise ValueError("Expected string after serialisation")
        return json.loads(jsonified)

    @classmethod
    def grid(cls, children: Children, **kwargs):
        """Create the responsive grid for a field."""
        return dmc.SimpleGrid(children, className="pydantic-form-grid " + kwargs.pop("className", ""), **kwargs)

    @classmethod
    def load(cls, **data) -> "FormLayout":
        """Load the form layout or a subclass."""
        adapter = TypeAdapter(Annotated[Union[tuple(get_all_subclasses(cls))], Discriminator("layout")])  # noqa: UP007
        return adapter.validate_python(data)



import itertools
from copy import deepcopy
from functools import partial
from typing import Literal

import dash_mantine_components as dmc
from dash import MATCH, ClientsideFunction, Input, Output, State, clientside_callback, dcc
from dash.development.base_component import Component
from dash_iconify import DashIconify

from dash_pydantic_form.form_section import FormSection
from dash_pydantic_form.i18n import _
from dash_pydantic_form.ids import form_base_id
from dash_pydantic_utils import deep_merge

from .form_layout import FormLayout, Position


class StepsFormLayout(FormLayout):
    """Steps form layout.

    Parameters
    ----------
    render_kwargs: dict | None
        Kwargs to pass to the render method.
    sections: list[FormSection]
        List of form sections.
    remaining_fields_position: Position
        Position of the remaining fields, one of 'top', 'bottom' or 'none'.
    """

    class ids:
        """Steps form ids."""

        steps = partial(form_base_id, "_pydf-steps")
        steps_save = partial(form_base_id, "_pydf-steps-save")
        steps_next = partial(form_base_id, "_pydf-steps-next")
        steps_previous = partial(form_base_id, "_pydf-steps-previous")
        steps_nsteps = partial(form_base_id, "_pydf-steps-nsteps")

    sections: list[FormSection]
    remaining_fields_position: Position = "top"
    layout: Literal["steps"] = "steps"

    def render(  # noqa: PLR0913
        self,
        *,
        field_inputs: dict[str, Component],
        aio_id: str,
        form_id: str,
        path: str,
        **_kwargs,
    ):
        """Render the sections in an accordion."""
        kwargs = deepcopy(self.render_kwargs or {})
        additional_steps = kwargs.pop("additional_steps", [])
        stepper_styles = deep_merge(
            {
                "root": {"display": "flex", "gap": "1.5rem", "padding": "0.75rem 0 3rem"},
                "content": {"flex": 1, "padding": 0},
                "steps": {"minWidth": 180},
                "step": {"cursor": "pointer"},
                "stepBody": {"padding-top": "0.6875rem"},
                "stepCompletedIcon": {"&>svg": {"width": 12}},
            },
            kwargs.pop("styles", {}),
        )
        stepper = dmc.Stepper(
            id=self.ids.steps(aio_id, form_id, path),
            active=0,
            orientation=kwargs.pop("orientation", "vertical"),
            size=kwargs.pop("size", "sm"),
            styles=stepper_styles,
            children=[
                dmc.StepperStep(
                    label=section.name,
                    icon=DashIconify(icon=section.icon) if section.icon else None,
                    children=self.grid([field_inputs[field] for field in section.fields if field in field_inputs]),
                )
                for section in self.sections
            ]
            + additional_steps,
            **kwargs,
        )

        steps = dmc.Box(
            [
                stepper,
                dmc.Group(
                    [
                        dmc.Button(
                            _("Back"),
                            id=self.ids.steps_previous(aio_id, form_id, path),
                            disabled=True,
                            size="compact-sm",
                            leftSection=DashIconify(icon="carbon:arrow-left", height=16),
                        ),
                        dmc.Button(
                            _("Next"),
                            id=self.ids.steps_next(aio_id, form_id, path),
                            size="compact-sm",
                            rightSection=DashIconify(icon="carbon:arrow-right", height=16),
                        ),
                    ],
                    style={
                        "position": "absolute",
                        "top": f"calc({78 * (len(self.sections) + len(additional_steps))}px + 1rem)",
                    },
                ),
                dcc.Store(data=len(self.sections), id=self.ids.steps_nsteps(aio_id, form_id, path)),
            ],
            pos="relative",
        )

        fields_not_in_sections = set(field_inputs) - set(itertools.chain.from_iterable(s.fields for s in self.sections))

        if self.remaining_fields_position == "none" or not fields_not_in_sections:
            return [steps]
        if self.remaining_fields_position == "top":
            return [self.grid([v for k, v in field_inputs.items() if k in fields_not_in_sections], mb="sm"), steps]
        return [steps, self.grid([v for k, v in field_inputs.items() if k in fields_not_in_sections], mb="sm")]


clientside_callback(
    ClientsideFunction(namespace="pydf", function_name="stepsDisable"),
    Output(StepsFormLayout.ids.steps_previous(MATCH, MATCH, MATCH), "disabled"),
    Output(StepsFormLayout.ids.steps_next(MATCH, MATCH, MATCH), "disabled"),
    Input(StepsFormLayout.ids.steps(MATCH, MATCH, MATCH), "active"),
    State(StepsFormLayout.ids.steps_nsteps(MATCH, MATCH, MATCH), "data"),
)

clientside_callback(
    ClientsideFunction(namespace="pydf", function_name="stepsPreviousNext"),
    Output(StepsFormLayout.ids.steps(MATCH, MATCH, MATCH), "active"),
    Input(StepsFormLayout.ids.steps_previous(MATCH, MATCH, MATCH), "n_clicks"),
    Input(StepsFormLayout.ids.steps_next(MATCH, MATCH, MATCH), "n_clicks"),
    State(StepsFormLayout.ids.steps(MATCH, MATCH, MATCH), "active"),
    State(StepsFormLayout.ids.steps_nsteps(MATCH, MATCH, MATCH), "data"),
    prevent_inital_call=True,
)



from collections.abc import Mapping
from functools import partial
from typing import ClassVar

import dash_mantine_components as dmc
from dash import MATCH, ClientsideFunction, Input, Output, State, clientside_callback, dcc
from dash.development.base_component import Component
from pydantic import BaseModel, Field, field_validator, model_validator
from pydantic.fields import FieldInfo

from dash_pydantic_form import ids as common_ids
from dash_pydantic_form.fields.base_fields import BaseField
from dash_pydantic_utils import (
    Quantity,
    get_fullpath,
    get_non_null_annotation,
)


class QuantityField(BaseField):
    """Quantity field."""

    unit_options: list[str] | Mapping[str, str] = Field(
        title="Unit options",
        description="List of units to show in the dropdown, if a Mapping, will use the values as unit labels",
    )
    auto_convert: bool | None = Field(
        title="Auto-convert between units when switching.",
        description="If None, checks whether auto-conversion is possible but falls back to False if impossible",
        default=None,
    )
    unit_select_width: str = Field(
        title="Width of the unit select",
        default="5rem",
    )

    kwargs_for_both: ClassVar[tuple[str, ...]] = ("size",)
    base_component = dmc.NumberInput

    class ids(BaseField.ids):
        """Quantity field ids."""

        conversion_store = partial(common_ids.field_dependent_id, "_pydf-quantity-conversions")

    @field_validator("unit_options")
    @classmethod
    def check_unit_options(cls, unit_options: list[str]):
        """Ensure unit_options has at least one value."""
        if len(unit_options) == 0:
            raise ValueError("At least one unit option must be defined.")
        return unit_options

    @model_validator(mode="after")
    def check_field(self):
        """Check auto convert field based on unit options."""
        if self.auto_convert is False:
            return self

        if len(self.unit_options) == 1:
            self.auto_convert = False
        else:
            try:
                units_info = [Quantity.get_unit_info(unit)[0] for unit in self.unit_options]
                if len(set(units_info)) != 1:
                    raise ValueError("Incompatible units found")
                self.auto_convert = True
            except ValueError:
                if self.auto_convert is True:
                    raise
                self.auto_convert = False

        return self

    def _render(  # noqa: PLR0913
        self,
        *,
        item: BaseModel,
        aio_id: str,
        form_id: str,
        field: str,
        parent: str = "",
        field_info: FieldInfo,
    ) -> Component:
        """Render the quantity field."""
        ann = field_info.annotation
        if ann is None:
            raise ValueError("field_info.annotation is None")
        ann = get_non_null_annotation(ann)
        if not issubclass(ann, BaseModel) or set(ann.model_fields) != {"unit", "value"}:
            raise TypeError("Quantity field should be used with a field that has value and unit sub-fields.")

        value = self.get_value(item, field, parent)

        # Try converting if the given uit does not match any of the input units
        if value and value.unit not in self.unit_options:
            value = Quantity(value.value, value.unit).to(next(iter(self.unit_options)))
        meta = "auto-convert" if self.auto_convert else "default"
        new_parent = get_fullpath(parent, field)
        value_id = common_ids.value_field(aio_id, form_id, "value", parent=new_parent, meta=meta)
        unit_id = common_ids.value_field(aio_id, form_id, "unit", parent=new_parent, meta=meta)

        current_unit = value.unit if value else next(iter(self.unit_options))
        with_select = len(self.unit_options) > 1 and not self.read_only
        input_kwargs = self.input_kwargs

        # Add unit suffix and placeholder if not already user-defined
        if not with_select:
            current_unit_label = (
                current_unit if isinstance(self.unit_options, list) else self.unit_options[current_unit]
            )
            if "suffix" not in input_kwargs and "prefix" not in input_kwargs:
                input_kwargs["suffix"] = f" {current_unit_label}"
            if "placeholder" not in input_kwargs:
                input_kwargs["placeholder"] = current_unit_label

        select_kwargs = {k: v for k, v in input_kwargs.items() if k in self.kwargs_for_both}

        return dmc.Group(
            [
                dmc.NumberInput(
                    label=self.get_title(field_info=field_info, field_name=field),
                    description=self.get_description(field_info),
                    value=value.value if value else None,
                    required=self.is_required(field_info),
                    readOnly=self.read_only,
                    id=value_id,
                    hideControls=True,
                    flex="1 1 10rem",
                    styles={"input": {"borderRadius": "var(--input-radius) 0 0 var(--input-radius)"}}
                    if with_select
                    else None,
                    **input_kwargs,
                    **{"data-currentunit": current_unit},
                ),
                dmc.Select(
                    value=current_unit,
                    id=unit_id,
                    data=self.unit_options
                    if isinstance(self.unit_options, list)
                    else [{"value": unit, "label": label} for unit, label in self.unit_options.items()],
                    ml=-1,
                    checkIconPosition="right",
                    readOnly=self.read_only,
                    flex=f"0 1 {self.unit_select_width}",
                    styles={"input": {"borderRadius": "0 var(--input-radius) var(--input-radius) 0"}},
                    display=None if with_select else "none",
                    **select_kwargs,
                ),
                dcc.Store(
                    id=self.ids.conversion_store(aio_id, form_id, "", new_parent, meta),
                    data={unit: list(Quantity.get_unit_info(unit)[1]) for unit in self.unit_options}
                    if self.auto_convert
                    else None,
                ),
            ],
            align="end",
            wrap="nowrap",
            gap=0,
        )

    def get_value(self, item: BaseModel, field: str, parent: str) -> Quantity | None:
        """Get the value of a model (parent, field) pair. Defined to allow overriding."""
        value = super().get_value(item, field, parent)
        if isinstance(value, Quantity):
            return value
        if isinstance(value, dict):
            return Quantity(**value)
        if value is None:
            return None
        raise TypeError(f"Expected Quantity, got {type(value)}")


clientside_callback(
    ClientsideFunction(namespace="pydf", function_name="convertQuantityUnit"),
    Output(common_ids.value_field(MATCH, MATCH, "value", MATCH, "auto-convert"), "value", allow_duplicate=True),
    Output(common_ids.value_field(MATCH, MATCH, "value", MATCH, "auto-convert"), "data-currentunit"),
    Input(common_ids.value_field(MATCH, MATCH, "unit", MATCH, "auto-convert"), "value"),
    State(common_ids.value_field(MATCH, MATCH, "value", MATCH, "auto-convert"), "value"),
    State(common_ids.value_field(MATCH, MATCH, "value", MATCH, "auto-convert"), "data-currentunit"),
    State(QuantityField.ids.conversion_store(MATCH, MATCH, "", MATCH, "auto-convert"), "data"),
    prevent_initial_call=True,
)



import warnings
from typing import TYPE_CHECKING, Literal

from dash_iconify import DashIconify
from pydantic import BaseModel

if TYPE_CHECKING:
    from dash_pydantic_form.form_layouts.form_layout import FormLayout

SectionRender = Literal["accordion", "tabs", "steps"]
Position = Literal["top", "bottom", "none"]


class FormSection(BaseModel):
    """Form section model.

    Parameters
    ----------
    name: str
        Section name.
    fields: list[str]
        List of field names from the pydantic model.
    icon: str | DashIconify | None
        Section icon displayed before the section name, optional.
    default_open: bool
        Whether the section is open by default, only used for accordion sections.
    """

    name: str
    fields: list[str]
    icon: str | DashIconify | None = None
    default_open: bool = False

    model_config = {"arbitrary_types_allowed": True}


def Sections(
    sections: list[FormSection],
    remaining_fields_position: Position = "top",
    render: SectionRender = "accordion",
    render_kwargs: dict | None = None,
) -> "FormLayout":
    """Adapter from Sections to FormLayouts for backward compatibility."""
    from dash_pydantic_form.form_layouts import FormLayout

    warnings.warn("Sections is deprecated, use FormLayouts instead.", DeprecationWarning, stacklevel=1)
    return FormLayout.load(
        sections=sections,
        remaining_fields_position=remaining_fields_position,
        layout=render,
        render_kwargs=render_kwargs,
    )



import math
import re
from functools import partial
from typing import Literal

import dash_mantine_components as dmc
import fsspec
from dash import ALL, MATCH, ClientsideFunction, Input, Output, State, callback, clientside_callback, ctx, dcc, html
from dash.development.base_component import Component
from dash_iconify import DashIconify
from pydantic import BaseModel
from pydantic.fields import FieldInfo

from dash_pydantic_form.fields.base_fields import BaseField
from dash_pydantic_form.i18n import _
from dash_pydantic_form.ids import field_dependent_id, value_field

PathType = Literal["file", "directory", "glob"]


class PathField(BaseField):
    """Path field."""

    path_type: PathType = "file"
    backend: str
    prefix: str = ""
    with_copy_button: bool = True
    modal_max_width: int = 800
    value_includes_prefix: bool = False
    page_size: int = 20

    class ids(BaseField.ids):
        """Model list field ids."""

        modal = partial(field_dependent_id, "_pydf-path-field-modal")
        modal_btn = partial(field_dependent_id, "_pydf-path-field-modal-btn")
        modal_btn_text = partial(field_dependent_id, "_pydf-path-field-modal-btn-text")
        filetree = partial(field_dependent_id, "_pydf-path-field-filetree")
        config = partial(field_dependent_id, "_pydf-path-field-config")
        prefix = partial(field_dependent_id, "_pydf-path-field-prefix")
        glob = partial(field_dependent_id, "_pydf-path-field-glob")
        pagination = partial(field_dependent_id, "_pydf-path-field-pagination")
        pagination_store = partial(field_dependent_id, "_pydf-path-field-pagination-store")
        filter = partial(field_dependent_id, "_pydf-path-field-filter")
        nav = lambda aio_id, form_id, field, parent, meta, path: field_dependent_id(  # noqa: E731
            "_pydf-path-field-nav", aio_id, form_id, field, parent, meta
        ) | {"path": path}
        checkbox = lambda aio_id, form_id, field, parent, meta, path: field_dependent_id(  # noqa: E731
            "_pydf-path-field-checkbox", aio_id, form_id, field, parent, meta
        ) | {"path": path}

    def fs(self) -> fsspec.AbstractFileSystem:
        """Get the filesystem."""
        return fsspec.filesystem(self.backend)

    def _render(  # noqa: PLR0913
        self,
        *,
        item: BaseModel,
        aio_id: str,
        form_id: str,
        field: str,
        parent: str = "",
        field_info: FieldInfo,
    ) -> Component:
        """Create a form field of type image to interact with the model field."""
        value = self.get_value(item, field, parent)
        if self.read_only:
            return self._render_read_only(value, field, field_info)

        inputs = [
            dmc.Button(
                dmc.ScrollArea(
                    value
                    or dmc.Text(
                        _("Click to select a {path_type}").format(path_type=_(self.path_type)),
                        size="xs",
                        fs="italic",
                    ),
                    id=self.ids.modal_btn_text(aio_id, form_id, field, parent),
                    offsetScrollbars="x",
                    scrollbars="x",
                    scrollbarSize="0.25rem",
                    mb="-0.125rem",
                ),
                size="sm",
                id=self.ids.modal_btn(aio_id, form_id, field, parent),
                variant="outline",
                color="dimmed",
                styles={
                    "inner": {"justifyContent": "start"},
                    "label": {"fontWeight": "normal"},
                    "root": {
                        "borderRadius": "0.25rem",
                        "border": "1px solid var(--mantine-color-default-border)",
                    },
                },
                px="0.5rem",
                flex=1,
            ),
        ]
        if self.path_type == "glob":
            if value:
                base_path_match = re.findall(r"^([^*]*)\/.*\*", value)
                if not base_path_match:
                    raise ValueError("Invalid glob pattern.")

                base_path = base_path_match[0]
                glob = value[len(base_path) :]
            else:
                base_path = None
                glob = ""

            if not value:
                inputs[0].children.children = _("Click to select a directory")
            inputs[0].styles["root"]["borderRadius"] = "0.25rem 0 0 0.25rem"

            inputs += [
                dmc.TextInput(
                    value=glob,
                    size="sm",
                    id=self.ids.glob(aio_id, form_id, field, parent),
                    styles={"input": {"borderRadius": "0 0.25rem 0.25rem 0"}},
                    debounce=350,
                    placeholder=_("Enter glob pattern"),
                    flex=1,
                    ml=-1,
                ),
            ]

        title = self.get_title(field_info, field_name=field)
        description = self.get_description(field_info)
        title_children: list = [title] + [
            html.Span(" *", style={"color": "var(--input-asterisk-color, var(--mantine-color-error))"})
        ] * self.is_required(field_info)

        return dmc.Stack(
            (title is not None)
            * [
                dmc.Text(
                    title_children,
                    size="sm",
                    mt=3,
                    mb=5,
                    fw=500,
                    lh=1.55,
                )
            ]
            + (title is not None and description is not None)
            * [dmc.Text(description, size="xs", c="dimmed", mt=-5, mb=5, lh=1.2)]
            + [self._base_group(value, inputs, (aio_id, form_id, field, parent))],
            gap=0,
        )

    def _base_group(self, value, inputs: list, id_args: tuple) -> Component:
        prefix = self.prefix.rstrip("/") + "/"
        return dmc.Group(
            [
                *inputs,
                dmc.Text(
                    dcc.Clipboard(target_id=value_field(*id_args), className="hover-clipboard"),
                    ml="0.5rem",
                    display=None if self.with_copy_button else "none",
                    flex=0,
                ),
                dmc.TextInput(
                    id=value_field(*id_args),
                    display="none",
                    value=value,
                ),
                dmc.Modal(
                    id=self.ids.modal(*id_args),
                    title=_("Select a {path_type}").format(
                        path_type=_(self.path_type if self.path_type != "glob" else "directory")
                    ),
                    size=f"min({self.modal_max_width}px, 100vw - 4rem)",
                    styles={"header": {"paddingTop": "0.75rem", "paddingBottom": "0.25rem", "minHeight": "2.75rem"}},
                ),
                dcc.Store(
                    id=self.ids.config(*id_args),
                    data={
                        "path_type": self.path_type,
                        "backend": self.backend,
                        "prefix": prefix,
                        "value_includes_prefix": self.value_includes_prefix,
                        "page_size": self.page_size,
                    },
                ),
            ],
            gap=0,
            wrap="nowrap",
        )

    @staticmethod
    def _get_subpath(prefix: str, value: str | None, val: str, rm_value: bool = False, safe: bool = False):
        subpath = val[len(prefix.split("://")[-1]) :]
        if rm_value and value:
            subpath = subpath.removeprefix(value).lstrip("/")
        if safe:
            subpath = subpath.replace(".", "||")

        return subpath

    @staticmethod
    def _base_button(children, icon: DashIconify | None = None, **kwargs):
        return dmc.Button(
            children,
            size="compact-sm",
            color="gray",
            variant="subtle",
            leftSection=icon,
            styles={"inner": {"justifyContent": "start"}, "label": {"fontWeight": "normal"}},
            **kwargs,
        )

    @staticmethod
    def _file_group(val):
        return dmc.Group(
            [DashIconify(icon="flat-color-icons:file", height=16), dmc.Text(val, size="sm")],
            gap="xs",
        )

    @staticmethod
    def _selectable_group(checkbox_id: dict, button: dmc.Button | dmc.Group | None = None, label=None):
        return dmc.Group(
            [
                dmc.Checkbox(
                    id=checkbox_id,
                    label=label,
                    size="sm",
                    style={"--checkbox-size": "1rem"},
                    pl=6.333,
                    styles={
                        "body": {"alignItems": "center"},
                        "label": {"paddingLeft": "0.625rem", "cursor": "pointer"},
                    },
                )
            ]
            + ([button] if button is not None else []),
            gap="0.625rem",
        )

    @classmethod
    def _breadcrumbs_group(  # noqa: PLR0913
        cls, prefix: str, value: str, id_parts: list[str], parts: list[str], current_filter: str | None
    ):
        return dmc.Group(
            [
                dmc.Breadcrumbs(
                    [
                        dmc.Tooltip(
                            cls._base_button(
                                DashIconify(icon="flat-color-icons:opened-folder", height=16),
                                id=cls.ids.nav(*id_parts[:-1], path="", meta="breadcrumbs"),
                            ),
                            label=prefix.rstrip("/"),
                        ),
                    ]
                    + (
                        [
                            cls._base_button(
                                part,
                                id=cls.ids.nav(
                                    *id_parts[:-1], path="/".join(parts[: i + 1]).replace(".", "||"), meta="breadcrumbs"
                                ),
                            )
                            for i, part in enumerate(parts)
                        ]
                        if value
                        else []
                    )
                    + [
                        dmc.TextInput(
                            size="xs",
                            placeholder=_("Filter with prefix"),
                            variant="unstyled",
                            id=cls.ids.filter(*id_parts),
                            classNames={"input": "path-field-filter-input"},
                            leftSection=DashIconify(icon="carbon:filter", height=14),
                            debounce=350,
                            value=current_filter,
                        )
                    ],
                    separator="/",
                    separatorMargin="0.125rem",
                    ml=-3,
                ),
            ],
            gap="0.125rem",
            pb="0.5rem",
            mb="0.5rem",
            pos="sticky",
            top="2.75rem",
            bg="var(--mantine-color-body)",
            style={
                "zIndex": 1,
                "borderBottom": "1px solid color-mix(in srgb, var(--mantine-color-default-border), #0000 50%)",
            },
            w="100%",
        )

    @classmethod
    def filetree(  # noqa: PLR0913
        cls,
        *,
        prefix: str,
        value: str,
        id_parts: list[str],
        parts: list[str],
        path_type: PathType,
        filtered_vals: list[dict],
        page_size: int,
        page: int,
        current_filter: str | None = None,
    ):
        """File tree."""
        return [
            cls._breadcrumbs_group(prefix, value, id_parts, parts, current_filter),
            *[
                cls._base_button(
                    cls._get_subpath(prefix, value, v["name"], rm_value=True),
                    icon=DashIconify(icon="flat-color-icons:folder", height=16),
                    id=cls.ids.nav(*id_parts, cls._get_subpath(prefix, value, v["name"], safe=True)),
                )
                if v["type"] != path_type
                else cls._selectable_group(
                    cls.ids.checkbox(*id_parts, cls._get_subpath(prefix, value, v["name"], safe=True)),
                    **{
                        "button": cls._base_button(
                            cls._get_subpath(prefix, value, v["name"], True),
                            icon=DashIconify(icon="flat-color-icons:folder", height=16),
                            id=cls.ids.nav(*id_parts, cls._get_subpath(prefix, value, v["name"], safe=True)),
                        )
                        if v["type"] == "directory"
                        else None,
                        "label": None
                        if v["type"] == "directory"
                        else cls._file_group(cls._get_subpath(prefix, value, v["name"], True)),
                    },
                )
                for v in filtered_vals[page_size * (page - 1) : page_size * page]
            ],
            *(
                [
                    dmc.Pagination(
                        total=math.ceil(len(filtered_vals) / page_size),
                        value=page,
                        id=cls.ids.pagination(*id_parts),
                        mt="0.5rem",
                        size="sm",
                        boundaries=2,
                        siblings=2,
                    ),
                    dcc.Store(data=value, id=cls.ids.pagination_store(*id_parts)),
                ]
                if len(filtered_vals) > page_size
                else []
            ),
            *(
                [
                    dmc.Checkbox(
                        id=PathField.ids.checkbox(*id_parts, value.replace(".", "||")),
                        label=_("Select {path_type}").format(path_type=_(path_type)),
                    )
                ]
                if len(filtered_vals) == 0 and not current_filter
                else []
            ),
            *(
                [dmc.Text(_("No match"), size="sm", c="dimmed", fs="italic", px="0.5rem")]
                if len(filtered_vals) == 0 and current_filter
                else []
            ),
        ]


# dmc.Skeleton
clientside_callback(
    ClientsideFunction(namespace="pydf", function_name="showPathFieldSkeletons"),
    Output(PathField.ids.modal(MATCH, MATCH, MATCH, MATCH, MATCH), "opened"),
    Output(PathField.ids.modal(MATCH, MATCH, MATCH, MATCH, MATCH), "children"),
    Input(PathField.ids.modal_btn(MATCH, MATCH, MATCH, MATCH, MATCH), "n_clicks"),
)


@callback(
    Output(PathField.ids.filetree(MATCH, MATCH, MATCH, MATCH), "children"),
    Input(PathField.ids.filetree(MATCH, MATCH, MATCH, MATCH), "id"),
    Input(PathField.ids.nav(MATCH, MATCH, MATCH, MATCH, ALL, ALL), "n_clicks"),
    Input(PathField.ids.pagination(MATCH, MATCH, MATCH, MATCH, ALL), "value"),
    Input(PathField.ids.filter(MATCH, MATCH, MATCH, MATCH, ALL), "value"),
    State(value_field(MATCH, MATCH, MATCH, MATCH), "value"),
    State(PathField.ids.config(MATCH, MATCH, MATCH, MATCH), "data"),
    State(PathField.ids.pagination_store(MATCH, MATCH, MATCH, MATCH, ALL), "data"),
)
def update_filetree(  # noqa: PLR0913
    id_, _navs, page, filter_str, value, config, pagination_value
):
    """Update the file tree."""
    fs = fsspec.filesystem(config["backend"])
    path_type = config["path_type"]
    prefix = config["prefix"]
    id_parts: list[str] = [id_["aio_id"], id_["form_id"], id_["field"], id_["parent"], id_["meta"]]
    value = (value or "").removeprefix(prefix)
    page = page[0] if page else 1
    pagination_value = pagination_value[0] if pagination_value else value

    from_filter = False
    if ctx.triggered_id and isinstance(ctx.triggered_id, dict):
        if "nav" in ctx.triggered_id["component"]:
            value = ctx.triggered_id["path"].replace("||", ".")
            page = 1
        elif "filter" in ctx.triggered_id["component"]:
            value = next(
                inp["id"]["path"].replace("||", ".")
                for inp in ctx.inputs_list[1][::-1]
                if inp["id"]["meta"] == "breadcrumbs"
            )
            from_filter = True
            page = 1
        else:
            value = pagination_value
    elif path_type == "glob":
        base_path_match = re.findall(r"^([^*]*)\/.*\*", value)
        value = value if not base_path_match else base_path_match[0]

    # This happens when the modal first opens after clicking on the button
    # in this case, we want to show the content one level up.
    if not ctx.triggered_id:
        val_split = value.rsplit("/", maxsplit=1)
        value = val_split[0]
        if len(val_split) > 1:
            from_filter = True
            filter_str = [val_split[1]]

    if path_type == "glob":
        path_type = "directory"

    refresh_ls = isinstance(ctx.triggered_id, dict) and ctx.triggered_id.get("component") == "_pydf-path-field-nav"
    vals: list[dict] = fs.ls(f"{prefix.rstrip('/')}/{(value or '').lstrip('/')}", detail=True, refresh=refresh_ls)

    filtered_vals = []
    for val in vals:
        subpath = PathField._get_subpath(prefix, value, val["name"])
        subpath2 = PathField._get_subpath(prefix, value, val["name"], rm_value=True)
        if (
            ((path_type == "file") or (val["type"] != "file"))
            and subpath != value
            and (not from_filter or not filter_str or not filter_str[0] or subpath2.startswith(filter_str[0]))
        ):
            filtered_vals.append(val)

    parts = value.split("/")
    page_size = config.get("page_size", 20)

    return PathField.filetree(
        prefix=prefix,
        value=value,
        id_parts=id_parts,
        parts=parts,
        path_type=path_type,
        filtered_vals=filtered_vals,
        page_size=page_size,
        page=page,
        current_filter=(filter_str or [""])[0] if from_filter else "",
    )


clientside_callback(
    ClientsideFunction(namespace="pydf", function_name="updatePathFieldValue"),
    Output(PathField.ids.modal(MATCH, MATCH, MATCH, MATCH, MATCH), "opened", allow_duplicate=True),
    Output(value_field(MATCH, MATCH, MATCH, MATCH, MATCH), "value", allow_duplicate=True),
    Output(PathField.ids.modal_btn_text(MATCH, MATCH, MATCH, MATCH, MATCH), "children"),
    Input(PathField.ids.checkbox(MATCH, MATCH, MATCH, MATCH, MATCH, ALL), "checked"),
    Input(PathField.ids.glob(MATCH, MATCH, MATCH, MATCH, ALL), "value"),
    State(PathField.ids.config(MATCH, MATCH, MATCH, MATCH, MATCH), "data"),
    State(PathField.ids.modal_btn_text(MATCH, MATCH, MATCH, MATCH, MATCH), "children"),
    prevent_initial_call=True,
)
